// This file was generated by ecm_qt_declare_logging_category(): DO NOT EDIT!

#ifndef ECM_QLOGGINGCATEGORY_KIO_KAMERA_LOG_KIO_KAMERA_LOG_H
#define ECM_QLOGGINGCATEGORY_KIO_KAMERA_LOG_KIO_KAMERA_LOG_H

#include <QLoggingCategory>

Q_DECLARE_LOGGING_CATEGORY(KIO_KAMERA_LOG)


#endif
